# Copyright (c) 2019 SUSE LLC
#
# This file is part of ec2imgutils.
#
# ec2imgutils is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ec2imgutils is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ec2imgutils.ase.  If not, see <http://www.gnu.org/licenses/>.

import logging
import os
import random
import datetime

from tempfile import mkstemp, mkdtemp

from botocore.exceptions import ClientError

from ec2imgutils.ec2imgutils import EC2ImgUtils


class EC2Setup(EC2ImgUtils):
    """Class to prepare an Amazon EC2 account with all necessary resources"""

    def __init__(
            self,
            access_key,
            region,
            secret_key,
            session_token,
            log_level=logging.INFO,
            log_callback=None
    ):
        EC2ImgUtils.__init__(
            self,
            log_level=log_level,
            log_callback=log_callback
        )

        self.access_key = access_key
        self.region = region
        self.secret_key = secret_key
        self.session_token = session_token

        self.internet_gateway_id = ''
        self.key_pair_name = ''
        self.route_table_id = ''
        self.security_group_id = ''
        self.ssh_private_key_file = ''
        self.temp_dir = ''
        self.vpc_subnet_id = ''
        self.vpc_id = ''

    # ---------------------------------------------------------------------
    def clean_up(self):
        if self.key_pair_name:
            self._remove_upload_key_pair()
        if self.security_group_id:
            self._remove_security_group()
        if self.vpc_id:
            self._remove_vpc()

    # ---------------------------------------------------------------------
    def create_security_group(self, vpc_id=None):
        self.log.debug('Creating temporary security group')
        group_description = 'ec2uploadimg created %s' % datetime.datetime.now()
        if not vpc_id:
            vpc_id = self.vpc_id
        # Avoid name collisions
        group_created = False
        response = None
        while not group_created:
            group_name = 'ec2uploadimg-%s' % (random.randint(1, 100))
            try:
                response = self._connect().create_security_group(
                    GroupName=group_name, Description=group_description,
                    VpcId=vpc_id
                )
            except ClientError as e:
                if not e.response['Error']['Code'] == 'InvalidGroup.Duplicate':
                    raise e
                # Generate a new group name and try again
                continue
            group_created = True
        self.security_group_id = response['GroupId']
        self.log.debug(
            'Temporary Security Group Created %s in vpc %s'
            % (self.security_group_id, vpc_id)
        )
        self._connect().authorize_security_group_ingress(
            GroupId=self.security_group_id,
            IpPermissions=[
                {'IpProtocol': 'tcp',
                 'FromPort': 22,
                 'ToPort': 22,
                 'IpRanges': [{'CidrIp': '0.0.0.0/0'}]}
            ])

        self.log.debug(
            "Successfully allowed incoming SSH port 22 for security "
            "group %s in %s" % (self.security_group_id, self.vpc_id)
        )
        return self.security_group_id

    # ---------------------------------------------------------------------
    def create_upload_key_pair(self, key_name='temporary_ec2_uploadkey'):
        self.log.debug('Creating temporary key pair')
        dir_path = os.path.expanduser('~/')
        if not os.access(dir_path, os.W_OK):
            dir_path = mkdtemp()
            self.temp_dir = dir_path
        fd, location = mkstemp(prefix='temporary_ec2_uploadkey.',
                               suffix='.key', dir=dir_path)
        self.key_pair_name = os.path.basename(location)
        self.ssh_private_key_file = location
        secret_key_content = self._connect().create_key_pair(
            KeyName=self.key_pair_name
        )
        self.log.debug(
            'Successfully created key pair: {}'.format(self.key_pair_name)
        )
        with open(location, 'w') as localfile:
            localfile.write(secret_key_content['KeyMaterial'])
        self.log.debug(
            'Successfully wrote secret key key file to {}'.format(location)
        )
        os.close(fd)
        return self.key_pair_name, self.ssh_private_key_file

    # ---------------------------------------------------------------------
    def create_vpc_subnet(self):
        self._create_vpc()
        self._create_internet_gateway()
        self._create_route_table()
        self._create_vpc_subnet()
        return self.vpc_subnet_id

    # ---------------------------------------------------------------------
    def _create_internet_gateway(self):
        response = self._connect().create_internet_gateway()
        self.internet_gateway_id = \
            response['InternetGateway']['InternetGatewayId']
        self._connect().attach_internet_gateway(
            VpcId=self.vpc_id, InternetGatewayId=self.internet_gateway_id
        )
        self.log.debug(
            "Successfully created internet gateway %s"
            % self.internet_gateway_id
        )

    # ---------------------------------------------------------------------
    def _create_route_table(self):
        response = self._connect().create_route_table(VpcId=self.vpc_id)
        self.route_table_id = response['RouteTable']['RouteTableId']
        self._connect().create_route(
            DestinationCidrBlock='0.0.0.0/0',
            GatewayId=self.internet_gateway_id,
            RouteTableId=self.route_table_id
        )
        self.log.debug(
            "Successfully created route table %s" % self.route_table_id
        )

    # ---------------------------------------------------------------------
    def _create_vpc(self):
        vpc_name = 'uploader-%s' % (random.randint(1, 100))
        response = self._connect().create_vpc(CidrBlock='192.168.0.0/16')
        self.vpc_id = response['Vpc']['VpcId']
        self._connect().create_tags(
            Resources=[self.vpc_id],
            Tags=[{'Key': 'Name', 'Value': vpc_name}]
        )
        self.log.debug("Successfully created VPC with id %s" % self.vpc_id)

    # ---------------------------------------------------------------------
    def _create_vpc_subnet(self):
        response = self._connect().create_subnet(
            CidrBlock='192.168.1.0/24', VpcId=self.vpc_id
        )
        self.vpc_subnet_id = response['Subnet']['SubnetId']
        self._connect().associate_route_table(
            SubnetId=self.vpc_subnet_id, RouteTableId=self.route_table_id
        )
        self._connect().modify_subnet_attribute(
            MapPublicIpOnLaunch={'Value': True}, SubnetId=self.vpc_subnet_id
        )
        self.log.debug(
            "Successfully created VPC subnet with id %s" % self.vpc_subnet_id
        )

    # ---------------------------------------------------------------------
    def _remove_security_group(self):
        self._connect().delete_security_group(
            GroupId=self.security_group_id
        )
        self.log.debug(
            'Successfully deleted security group %s' % self.security_group_id
        )

    # ---------------------------------------------------------------------
    def _remove_upload_key_pair(self):
        self.log.debug(
            'Deleting temporary key pair {}'.format(self.key_pair_name)
        )
        self._connect().delete_key_pair(
            KeyName=self.key_pair_name)
        if os.path.isfile(self.ssh_private_key_file):
            os.remove(self.ssh_private_key_file)
        if self.temp_dir:
            os.rmdir(self.temp_dir)
        self.log.debug(
            'Successfully deleted temporary key %s'
            % self.ssh_private_key_file
        )

    # ---------------------------------------------------------------------
    def _remove_vpc(self):
        self._connect().delete_route(
            DestinationCidrBlock='0.0.0.0/0', RouteTableId=self.route_table_id
        )
        self.log.debug(
            'Successfully deleted route from route table %s'
            % self.route_table_id
        )
        self._connect().delete_subnet(SubnetId=self.vpc_subnet_id)
        self.log.debug(
            'Successfully deleted VPC subnet %s' % self.vpc_subnet_id
        )
        self._connect().delete_route_table(RouteTableId=self.route_table_id)
        self.log.debug(
            'Successfully deleted route table %s' % self.route_table_id
        )
        self._connect().detach_internet_gateway(
            InternetGatewayId=self.internet_gateway_id, VpcId=self.vpc_id
        )
        self.log.debug(
            'Successfully deleted detached internet gateway %s'
            % self.internet_gateway_id
        )
        self._connect().delete_internet_gateway(
            InternetGatewayId=self.internet_gateway_id
        )
        self.log.debug(
            'Successfully deleted internet gateway %s'
            % self.internet_gateway_id
        )
        self._connect().delete_vpc(VpcId=self.vpc_id)
        self.log.debug('Successfully deleted VPC %s' % self.vpc_id)



# Copyright (c) 2018 SUSE LLC
#
# This file is part of ec2imgutils.
#
# ec2imgutils is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ec2imgutils is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ec2imgutils.ase.  If not, see <http://www.gnu.org/licenses/>.

import boto3
import logging

from ec2imgutils.ec2imgutilsExceptions import EC2ConnectionException


class EC2ImgUtils:
    """Base class for EC2 Image Utilities"""

    def __init__(self, log_level=logging.INFO, log_callback=None):

        self.region = None
        self.session_token = None

        if log_callback:
            self.log = log_callback
        else:
            logger = logging.getLogger('ec2imgutils')
            logger.setLevel(log_level)
            self.log = logger

        try:
            self.log_level = self.log.level
        except AttributeError:
            self.log_level = self.log.logger.level  # LoggerAdapter

    # ---------------------------------------------------------------------
    def _connect(self):
        """Connect to EC2"""

        ec2 = None
        if self.region:
            if self.session_token:
                ec2 = boto3.client(
                    aws_access_key_id=self.access_key,
                    aws_secret_access_key=self.secret_key,
                    aws_session_token=self.session_token,
                    region_name=self.region,
                    service_name='ec2'
                )
            else:
                session = boto3.session.Session()
                ec2 = session.client(
                    aws_access_key_id=self.access_key,
                    aws_secret_access_key=self.secret_key,
                    region_name=self.region,
                    service_name='ec2'
                )
        else:
            self.region = 'UNKNOWN'

        if not ec2:
            msg = 'Could not connect to region: %s ' % self.region
            raise EC2ConnectionException(msg)

        return ec2

    # ---------------------------------------------------------------------
    def _get_owned_images(self):
        """Return the list of images owned by the account used for
           uploading"""
        return self._connect().describe_images(Owners=['self'])['Images']

    # ---------------------------------------------------------------------
    def _set_access_keys(self):
        """Set the access keys for the connection"""
        if not self.access_key:
            self.access_key = self.config.get_option(self.account,
                                                     'access_key_id')
        if not self.secret_key:
            self.secret_key = self.config.get_option(self.account,
                                                     'secret_access_key')

    # ---------------------------------------------------------------------
    def set_region(self, region):
        """Set the region that should be used."""
        if self.region and self.region == region:
            return True

        self.log.debug('Using EC2 region: {}'.format(region))
        self.region = region

        return True



# Copyright 2021 SUSE LLC
#
# This file is part of ec2imgutils
#
# ec2imgutils is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ec2imgutils is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ec2deprecateimg. If not, see <http://www.gnu.org/licenses/>.

import datetime
import dateutil.relativedelta
import logging

import ec2imgutils.ec2utils as utils
from ec2imgutils.ec2imgutils import EC2ImgUtils
from ec2imgutils.ec2imgutilsExceptions import EC2DeprecateImgException


class EC2DeprecateImg(EC2ImgUtils):
    """Deprecate EC2 image(s) by tagging the image with 3 tags, Deprecated on,
       Removal date, and Replacement image."""

    def __init__(
            self,
            access_key=None,
            deprecation_period=6,
            deprecation_image_id=None,
            deprecation_image_name=None,
            deprecation_image_name_fragment=None,
            deprecation_image_name_match=None,
            force=None,
            image_virt_type=None,
            public_only=None,
            replacement_image_id=None,
            replacement_image_name=None,
            replacement_image_name_fragment=None,
            replacement_image_name_match=None,
            secret_key=None,
            log_level=logging.INFO,
            log_callback=None
    ):
        EC2ImgUtils.__init__(
            self,
            log_level=log_level,
            log_callback=log_callback
        )

        self.access_key = access_key
        self.deprecation_period = deprecation_period
        self.deprecation_image_id = deprecation_image_id
        self.deprecation_image_name = deprecation_image_name
        self.deprecation_image_name_fragment = deprecation_image_name_fragment
        self.deprecation_image_name_match = deprecation_image_name_match
        self.force = force
        self.image_virt_type = image_virt_type
        self.public_only = public_only
        self.replacement_image_id = replacement_image_id
        self.replacement_image_name = replacement_image_name
        self.replacement_image_name_fragment = replacement_image_name_fragment
        self.replacement_image_name_match = replacement_image_name_match
        self.secret_key = secret_key

        self._set_deprecation_date()
        self._set_deletion_date()

        self.replacement_image_tag = None

    # ---------------------------------------------------------------------
    def _find_images_by_id(self, image_id, filter_replacement_image=None):
        """Find images by ID match"""
        my_images = self._get_all_type_match_images(filter_replacement_image)
        return utils.find_images_by_id(my_images, image_id)

    # ---------------------------------------------------------------------
    def _find_images_by_name(self, image_name, filter_replacement_image=None):
        """Find images by exact name match"""
        my_images = self._get_all_type_match_images(filter_replacement_image)
        return utils.find_images_by_name(my_images, image_name, self.log)

    # ---------------------------------------------------------------------
    def _find_images_by_name_fragment(
            self,
            name_fragment,
            filter_replacement_image=None):
        """Find images by string matching of the fragment with the name"""
        my_images = self._get_all_type_match_images(filter_replacement_image)
        return utils.find_images_by_name_fragment(
            my_images,
            name_fragment,
            self.log
        )

    # ---------------------------------------------------------------------
    def _find_images_by_name_regex_match(
            self,
            expression,
            filter_replacement_image=None):
        """Find images by match the name with the given regular expression"""
        my_images = self._get_all_type_match_images(filter_replacement_image)
        return utils.find_images_by_name_regex_match(
            my_images,
            expression,
            self.log
        )

    # ---------------------------------------------------------------------
    def _format_date(self, date):
        """Format the date to YYYYMMDD"""

        return date.strftime('%Y%m%d')

    # ---------------------------------------------------------------------
    def _get_all_type_match_images(self, filter_replacement_image=None):
        """Get all images that match thespecified virtualization type.
           All images owned by the account if not type is specified."""
        images = []
        my_images = self._get_owned_images()
        for image in my_images:
            if filter_replacement_image:
                if image['ImageId'] == self.replacement_image_id:
                    msg = 'Ignore replacement image as potential target '
                    msg += 'for deprecation.'
                    self.log.debug(msg)
                    continue
            if self.image_virt_type:
                if self.image_virt_type == image['VirtualizationType']:
                    images.append(image)
                else:
                    continue
            if self.public_only:
                launch_attributes = self._connect().describe_image_attribute(
                    ImageId=image['ImageId'],
                    Attribute='launchPermission')['LaunchPermissions']
                launch_permission = None
                if launch_attributes:
                    launch_permission = launch_attributes[0].get('Group', None)
                if launch_permission == 'all':
                    images.append(image)
                continue
            images.append(image)

        return images

    # ---------------------------------------------------------------------
    def _get_images_to_deprecate(self):
        """Find images to deprecate"""
        images = None
        if self.deprecation_image_id:
            images = self._find_images_by_id(self.deprecation_image_id, True)
        elif self.deprecation_image_name:
            images = self._find_images_by_name(
                self.deprecation_image_name, True)
        elif self.deprecation_image_name_fragment:
            images = self._find_images_by_name_fragment(
                self.deprecation_image_name_fragment, True)
        elif self.deprecation_image_name_match:
            images = self._find_images_by_name_regex_match(
                self.deprecation_image_name_match, True)
        else:
            msg = 'No deprecation image condition set. Should not reach '
            msg += 'this point.'
            raise EC2DeprecateImgException(msg)

        return images

    # ---------------------------------------------------------------------
    def _set_deletion_date(self):
        """Set the date when the deprecation perios expires"""
        now = datetime.datetime.now()
        expire = now + dateutil.relativedelta.relativedelta(
            months=+self.deprecation_period)
        self.deletion_date = self._format_date(expire)

    # ---------------------------------------------------------------------
    def _set_deprecation_date(self):
        """Set the deprecation day in the YYYYMMDD format"""
        now = datetime.datetime.now()
        self.deprecation_date = self._format_date(now)

    # ---------------------------------------------------------------------
    def _set_replacement_image_info(self):
        """Find the replacement image information and create an identifier"""
        images = None
        condition = None
        if self.replacement_image_id:
            condition = self.replacement_image_id
            images = self._find_images_by_id(condition)
        elif self.replacement_image_name:
            condition = self.replacement_image_name
            images = self._find_images_by_name(condition)
        elif self.replacement_image_name_fragment:
            condition = self.replacement_image_name_fragment
            images = self._find_images_by_name_fragment(condition)
        elif self.replacement_image_name_match:
            condition = self.replacement_image_name_match
            images = self._find_images_by_name_regex_match(condition)
        else:
            msg = 'No replacement image condition set. Should not reach '
            msg += 'this point.'
            raise EC2DeprecateImgException(msg)

        if not images:
            msg = 'Replacement image not found, "%s" ' % condition
            msg += 'did not match any image.'
            raise EC2DeprecateImgException(msg)

        if len(images) > 1:
            msg = 'Replacement image ambiguity, the specified condition '
            msg += '"%s" return multiple replacement image options' % condition
            raise EC2DeprecateImgException(msg)

        image = images[0]
        self.replacement_image_tag = '%s -- %s' % (
            image['ImageId'],
            image['Name']
        )

    # ---------------------------------------------------------------------
    def deprecate_images(self):
        """Deprecate images in the connected region"""
        self._connect()
        self._set_replacement_image_info()
        images = self._get_images_to_deprecate()
        if not images:
            self.log.debug('No images to deprecate found')
            return False

        self.log.debug('Deprecating images in region: {}'.format(self.region))
        self.log.debug('\tDeprecated on {}'.format(self.deprecation_date))
        self.log.debug('Removal date {}'.format(self.deletion_date))
        self.log.debug(
            'Replacement image {}'.format(self.replacement_image_tag)
        )

        ec2 = self._connect()
        for image in images:
            existing_tags = image.get('Tags')
            tagged = False
            if not self.force and existing_tags:
                for tag in existing_tags:
                    if tag.get('Key') == 'Deprecated on':
                        msg = '\t\tImage %s already tagged, skipping'
                        self.log.debug(msg % image['ImageId'])
                        tagged = True
            if tagged:
                continue
            deprecated_on_data = {
                'Key': 'Deprecated on',
                'Value': self.deprecation_date
            }
            removal_date_data = {
                'Key': 'Removal date',
                'Value': self.deletion_date
            }
            replacement_image_data = {
                'Key': 'Replacement image',
                'Value': self.replacement_image_tag
            }
            tags = [
                deprecated_on_data,
                removal_date_data,
                replacement_image_data
            ]
            ec2.create_tags(
                Resources=[image['ImageId']], Tags=tags
            )
            self.log.debug(
                '\t\ttagged:%s\t%s' % (image['ImageId'], image['Name'])
            )
            # There is a difference in terminology. In EC2 deprecated means
            # an image cannot be used anymore by new users, this is similar
            # to a deletion. For ec2imgutils deprecation means an image should
            # no longer be used. Therefore we set the calculated deletion
            # date as the deprecation date in AWS
            ec2.enable_image_deprecation(
                ImageId=image['ImageId'],
                DeprecateAt=datetime.datetime.strptime(
                    self.deletion_date, '%Y%m%d'
                )
            )

    # ---------------------------------------------------------------------
    def print_deprecation_info(self):
        """Print information about the images that would be deprecated."""
        self._connect()
        self._set_replacement_image_info()
        images = self._get_images_to_deprecate()
        if not images:
            self.log.info('No images to deprecate found')
            return True

        self.log.info(
            'Would deprecate images in region: {}'.format(self.region)
        )
        self.log.info('\tDeprecated on {}'.format(self.deprecation_date))
        self.log.info('\tRemoval date {}'.format(self.deletion_date))
        self.log.info(
            '\tReplacement image {}'.format(self.replacement_image_tag)
        )
        self.log.info('\tImages to deprecate:\n\t\tID\t\t\t\tName')
        for image in images:
            self.log.info('\t\t%s\t%s' % (image['ImageId'], image['Name']))

        return True



# Copyright 2018 SUSE LLC
#
# This file is part of ec2imgutils
#
# ec2imgutils is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ec2imgutils is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ec2publishimg. If not, see <http://www.gnu.org/licenses/>.

import logging
import sys
import time

import ec2imgutils.ec2utils as utils
from ec2imgutils.ec2imgutils import EC2ImgUtils
from ec2imgutils.ec2imgutilsExceptions import EC2RemoveImgException


class EC2RemoveImage(EC2ImgUtils):
    """Remove images from EC2."""

    # --------------------------------------------------------------------
    def __init__(
            self,
            access_key=None,
            image_id=None,
            image_name=None,
            image_name_fragment=None,
            image_name_match=None,
            keep_snap=False,
            confirm=None,
            remove_all=False,
            secret_key=None,
            log_level=logging.INFO,
            log_callback=None
    ):
        EC2ImgUtils.__init__(
            self,
            log_level=log_level,
            log_callback=log_callback
        )

        self.access_key = access_key
        self.image_id = image_id
        self.image_name = image_name
        self.image_name_fragment = image_name_fragment
        self.image_name_match = image_name_match
        self.keep_snap = keep_snap
        self.confirm = confirm
        self.remove_all = remove_all
        self.secret_key = secret_key

    # ---------------------------------------------------------------------
    def _check_images_boundary_condition(self, images):
        """Check if the images found meet operating conditions:
           If all is true we delete all found images if not we should have
           found only one image"""

        if not images:
            self.log.info(
                'No images to remove found in region: {}'.format(self.region)
            )

        if len(images) > 1 and not self.remove_all:
            msg = 'Found multiple images to remove, but "all" is '
            msg += 'not set. Cannot disambiguate images to remove'
            self.log.info(msg)
            return False

        return True

    # ---------------------------------------------------------------------
    def _get_images_to_remove(self):
        """Find the images to remove"""
        owned_images = self._get_owned_images()
        if self.image_id:
            return utils.find_images_by_id(owned_images, self.image_id)
        elif self.image_name:
            return utils.find_images_by_name(
                owned_images,
                self.image_name,
                self.log
            )
        elif self.image_name_fragment:
            return utils.find_images_by_name_fragment(
                owned_images,
                self.image_name_fragment,
                self.log
            )
        elif self.image_name_match:
            try:
                return utils.find_images_by_name_regex_match(
                    owned_images,
                    self.image_name_match,
                    self.log
                )
            except Exception:
                msg = 'Unable to complie regular expression "%s"'
                msg = msg % self.image_name_match
                raise EC2RemoveImgException(msg)
        else:
            msg = 'No remove image condition set. Should not reach '
            msg += 'this point.'
            raise EC2RemoveImgException(msg)

    # ---------------------------------------------------------------------
    def _get_snapshot_id(self, image):
        """Get the snapshot id associated with this image"""
        device_map = image.get('BlockDeviceMappings')
        if not device_map:
            error_msg = 'Image "%s" has no device map' % image.get('Name')
            raise EC2RemoveImgException(error_msg)
        # The image snapshot should be the first device
        ebs_info = device_map[0].get('Ebs')
        if not ebs_info:
            error_msg = 'Image "%s" is not EBS backed' % image.get('Name')
            raise EC2RemoveImgException(error_msg)
        snapshot_id = ebs_info.get('SnapshotId')
        if not snapshot_id:
            error_msg = 'No snapshot found for image "%s"' % image.get('Name')
            raise EC2RemoveImgException(error_msg)
        return snapshot_id

    # ---------------------------------------------------------------------
    def print_remove_info(self):
        """Print information about images that would be deleted"""
        images = self._get_images_to_remove()
        images_ok = self._check_images_boundary_condition(images)

        if not images_ok:
            # While the boundary conditions are not met, we are just printing
            # information thus this is not considered a failure
            return True

        header_msg = 'Would remove image '
        if not self.keep_snap:
            header_msg += 'and snapshot '
        header_msg += 'in region: {}'
        self.log.info(header_msg.format(self.region))
        for image in images:
            if not self.keep_snap:
                snapshot = self._get_snapshot_id(image)
                self.log.info('\t\t%s\t%s\t%s' % (
                    image['ImageId'],
                    image['Name'],
                    snapshot
                ))
            else:
                self.log.info('\t\t%s\t%s' % (image['ImageId'], image['Name']))

        return True

    # ---------------------------------------------------------------------
    def remove_images(self):
        """Remove the images"""
        ec2 = self._connect()
        images = self._get_images_to_remove()
        images_ok = self._check_images_boundary_condition(images)

        if not images_ok:
            raise EC2RemoveImgException('Image ambiguity')

        for image in images:
            delete = 'True'
            if self.confirm:
                delete = self._query_yes_no(image)

            if delete:
                ec2.deregister_image(ImageId=image['ImageId'])
                if not self.keep_snap:
                    snapshot = self._get_snapshot_id(image)
                    # Give the EC2 backend a little bit of time to catch up
                    time.sleep(1)
                    ec2.delete_snapshot(SnapshotId=snapshot)

                    self.log.debug(
                        'Removing in region: {}'.format(self.region))
                    self.log.debug(
                        '\tImage: %s\t%s' % (image['ImageId'], image['Name'])
                    )
                if not self.keep_snap:
                    self.log.debug('\tSnapshot: {}'.format(snapshot))
            else:
                continue
    # ---------------------------------------------------------------------

    def _query_yes_no(self, image):
        yes = {'yes', 'y', 'ye', ''}
        no = {'no', 'n'}

        while True:
            try:
                choice = input(
                            '\tConfirm Delete Image: %s\t%s (Y/n)'
                            % (image['ImageId'], image['Name'])
                        ).lower()
                if choice in yes:
                    return True
                elif choice in no:
                    return False
                else:
                    sys.stdout.write("Please respond with 'yes' or 'no'\n")
            except KeyboardInterrupt:
                raise EC2RemoveImgException(
                        'Keyboard Interrupt received, exiting')



# Copyright 2020 SUSE LLC
#
# This file is part of ec2imgutils
#
# ec2imgutils is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ec2imgutils is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ec2publishimg. If not, see <http://www.gnu.org/licenses/>.

import logging

import ec2imgutils.ec2utils as utils
from ec2imgutils.ec2imgutils import EC2ImgUtils
from ec2imgutils.ec2imgutilsExceptions import EC2PublishImgException


class EC2PublishImage(EC2ImgUtils):
    """Publish/share EC2 image(s). Will also hide the image(s)"""

    # --------------------------------------------------------------------
    def __init__(
            self,
            access_key=None,
            allow_copy='none',
            image_id=None,
            image_name=None,
            image_name_fragment=None,
            image_name_match=None,
            secret_key=None,
            visibility='all',
            log_level=logging.INFO,
            log_callback=None
    ):
        EC2ImgUtils.__init__(
            self,
            log_level=log_level,
            log_callback=log_callback
        )

        self.access_key = access_key
        self.image_id = image_id
        self.image_name = image_name
        self.image_name_fragment = image_name_fragment
        self.image_name_match = image_name_match
        self.secret_key = secret_key
        self.visibility = visibility

        if self.visibility == 'all':
            self.publish_msg = '\tPublished: %s\t\t%s'
        elif self.visibility == 'none':
            self.publish_msg = '\tSet to private: %s\t\t%s'
        else:
            self.publish_msg = '\tShared: %s\t\t%s\t with: %s'

        if allow_copy == 'image':
            self.allow_copy = self.visibility
        else:
            self.allow_copy = allow_copy

    # --------------------------------------------------------------------
    def _get_images(self):
        """Return a list of images that match the filter criteria"""
        self._connect()
        owned_images = self._get_owned_images()
        if self.image_id:
            return utils.find_images_by_id(owned_images, self.image_id)
        elif self.image_name:
            return utils.find_images_by_name(
                owned_images,
                self.image_name,
                self.log
            )
        elif self.image_name_fragment:
            return utils.find_images_by_name_fragment(
                owned_images,
                self.image_name_fragment,
                self.log
            )
        elif self.image_name_match:
            try:
                return utils.find_images_by_name_regex_match(
                    owned_images,
                    self.image_name_match,
                    self.log
                )
            except Exception:
                msg = 'Unable to complie regular expression "%s"'
                msg = msg % self.image_name_match
                raise EC2PublishImgException(msg)

    # --------------------------------------------------------------------
    def _get_snapshot_ids_for_image(self, image):
        """Return the snapshot ID for a given image"""
        image_data = self._connect().describe_images(
            ImageIds=[image['ImageId']])['Images']
        block_device_maps = image_data[0]['BlockDeviceMappings']
        snapshot_ids = []
        for block_map in block_device_maps:
            snapshot_ids.append(block_map['Ebs']['SnapshotId'])

        return snapshot_ids

    # --------------------------------------------------------------------
    def _share_snapshot(self, image):
        """Provide permission to copy the underlying snapshot"""

        snapshot_ids = self._get_snapshot_ids_for_image(image)
        for snapshot_id in snapshot_ids:
            if self.allow_copy == 'all':
                self._connect().modify_snapshot_attribute(
                    SnapshotId=snapshot_id,
                    Attribute='createVolumePermission',
                    OperationType='add',
                    GroupNames=['all']
                )
            elif self.allow_copy != 'none':
                self._connect().modify_snapshot_attribute(
                    SnapshotId=snapshot_id,
                    Attribute='createVolumePermission',
                    OperationType='add',
                    UserIds=self.allow_copy.split(',')
                )

    # --------------------------------------------------------------------
    def _print_image_info(self, image, log_callback):
        """Print a message about the image that would be modified"""
        if self.visibility == 'all' or self.visibility == 'none':
            log_callback(
                self.publish_msg % (image['ImageId'], image['Name'])
            )
        else:
            log_callback(self.publish_msg % (
                        image['ImageId'],
                        image['Name'],
                        self.visibility
            ))

    # --------------------------------------------------------------------
    def print_publish_info(self):
        """Print information about images that would be published"""
        images = self._get_images()
        for image in images:
            self._print_image_info(image, log_callback=self.log.info)

    # --------------------------------------------------------------------
    def publish_images(self):
        """Publish the matching image(s)"""
        images = self._get_images()

        for image in images:
            skip_state = ['invalid', 'deregistered', 'transient', 'fail']
            if image['State'] == 'pending':
                waiter = self._connect().get_waiter('image_available')
                self.log.info("Current state of image %s is %s. Waiting up "
                              "to 10 minutes for it to become 'available'.",
                              image['ImageId'], image['State'])
                wait_status = 1
                try:
                    wait_status = waiter.wait(
                        ImageIds=[image['ImageId']],
                        Filters=[
                            {
                                'Name': 'state',
                                'Values': ['available']
                            }
                        ],
                        WaiterConfig={
                            'Delay': 15,
                            'MaxAttempts': 40
                        }
                    )
                except Exception:
                    wait_status = 1

                if wait_status:
                    self.log.info("Skipping image %s as it failed to become"
                                  "'available'", image['ImageId'])
                    continue
                else:
                    self.log.info("Image %s is now 'available'.",
                                  image['ImageId'])

            elif image['State'] in skip_state:
                self.log.info("Skipping image %s as the state is %s"
                              " and is not able to be published.",
                              image['ImageId'], image['State'])
                continue

            if self.visibility == 'all':
                self._connect().modify_image_attribute(
                    ImageId=image['ImageId'],
                    Attribute='launchPermission',
                    OperationType='add',
                    UserGroups=['all']
                )
                if self.allow_copy != 'none':
                    self._share_snapshot(image)
            elif self.visibility == 'none':
                launch_attributes = self._connect().describe_image_attribute(
                    ImageId=image['ImageId'],
                    Attribute='launchPermission'
                )['LaunchPermissions']
                launch_permission = {
                    'Remove': launch_attributes
                }
                if not launch_attributes:
                    msg = '\tImage with ID: %s  ' % image['ImageId']
                    msg += 'is already private, nothing to do'
                    self.log.info(msg)
                    continue
                self._connect().modify_image_attribute(
                    ImageId=image['ImageId'],
                    LaunchPermission=launch_permission
                )
                snapshot_ids = self._get_snapshot_ids_for_image(image)
                for snapshot_id in snapshot_ids:
                    snapshot_attrs = (
                        self._connect().describe_snapshot_attribute(
                            SnapshotId=snapshot_id,
                            Attribute='createVolumePermission'
                        )['CreateVolumePermissions']
                    )
                    if not snapshot_attrs:
                        continue
                    snapshot_permission = {
                        'Remove': snapshot_attrs
                    }
                    self._connect().modify_snapshot_attribute(
                        SnapshotId=snapshot_id,
                        CreateVolumePermission=snapshot_permission
                    )
            else:
                self._connect().modify_image_attribute(
                    ImageId=image['ImageId'],
                    Attribute='launchPermission',
                    OperationType='add',
                    UserIds=self.visibility.split(',')
                )
                if self.allow_copy:
                    self._share_snapshot(image)

            self._print_image_info(image, log_callback=self.log.debug)



# Copyright (c) 2018 SUSE LLC
#
# This file is part of ec2imgutils
#
# ec2imgutils is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ec2imgutils is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ec2imgutils.ase.  If not, see <http://www.gnu.org/licenses/>.

import boto3
import configparser
import logging
import os
import re
import sys

from itertools import repeat

from ec2imgutils.ec2imgutilsExceptions import (
    EC2AccountException,
    EC2ConfigFileParseException
)


# ----------------------------------------------------------------------------
def find_images_by_id(images, image_id):
    """Return a list of images that match the given ID. By definition this
       is a list of one as IDs are unique."""
    matching_images = []
    for image in images:
        if image_id == image['ImageId']:
            matching_images.append(image)
            # The framework guarantees unique image IDs
            break

    return matching_images


# ----------------------------------------------------------------------------
def find_images_by_name(images, image_name, log_callback):
    """Return a list of images that match the given name."""
    matching_images = []
    for image in images:
        if not image.get('Name'):
            _no_name_warning(image, log_callback)
            continue
        if image_name == image['Name']:
            matching_images.append(image)

    return matching_images


# ----------------------------------------------------------------------------
def find_images_by_name_fragment(images, image_name_fragment, log_callback):
    """Return a list of images that match the given fragment in any part
       of the image name."""
    matching_images = []
    for image in images:
        if not image.get('Name'):
            _no_name_warning(image, log_callback)
            continue
        if image['Name'].find(image_name_fragment) != -1:
            matching_images.append(image)

    return matching_images


# ----------------------------------------------------------------------------
def find_images_by_name_regex_match(images, image_name_regex, log_callback):
    """Return a list of images that match the given regular expression in
       their name."""
    matching_images = []
    image_name_exp = re.compile(image_name_regex)
    for image in images:
        if not image.get('Name'):
            _no_name_warning(image, log_callback)
            continue
        if image_name_exp.match(image['Name']):
            matching_images.append(image)

    return matching_images


# -----------------------------------------------------------------------------
def generate_config_account_name(account):
    """Generate the name of an account as it expected in the configuration"""
    return 'account-%s' % account


# -----------------------------------------------------------------------------
def generate_config_region_name(region):
    """Generate the name for a region as it is expected in the configuration"""
    return 'region-%s' % region


# -----------------------------------------------------------------------------
def get_config(configFilePath):
    """Return a config object fr hte given file."""

    config = configparser.RawConfigParser()
    parsed = None
    try:
        parsed = config.read(configFilePath)
    except Exception:
        msg = 'Could not parse configuration file "%s"\n' % configFilePath
        e_type, value, tb = sys.exc_info()
        msg += format(value)
        raise EC2ConfigFileParseException(msg)

    if not parsed:
        msg = 'Error parsing config file: %s' % configFilePath
        raise EC2ConfigFileParseException(msg)

    return config


# -----------------------------------------------------------------------------
def get_account_info_from_aws(account, entry):
    """Return a access value from the aws credentials file."""

    config = configparser.RawConfigParser()
    parsed = None
    value = None
    configFilePath = ['~/.aws/credentials', '~/.aws/config']

    for config_file in configFilePath:
        aws_file = os.path.expanduser(config_file)

        try:
            parsed = config.read(aws_file)
        except Exception:
            continue

        if not parsed:
            msg = 'Error parsing config file: %s' % aws_file
            raise EC2ConfigFileParseException(msg)

        try:
            item = 'aws_' + entry
            if config_file == '~/.aws/config':
                section = 'profile ' + account
            else:
                section = account

            if config.has_option(section, item):
                value = config.get(section, item)

        except Exception:
            continue
    if not value:
        raise EC2AccountException

    return value


# -----------------------------------------------------------------------------
def get_from_config(account, config, region, entry, cmd_line_arg):
    """Retrieve an entry from the configuration"""
    value = None
    if region:
        region_name = generate_config_region_name(region)
        # Region config over rides default account configuration
        if config.has_option(region_name, entry):
            value = config.get(region_name, entry)

    if not value:
        if not account:
            msg = 'No account given; missing command line argument %s'
            raise EC2AccountException(msg % cmd_line_arg)

        account_name = generate_config_account_name(account)

        if config.has_option(account_name, entry):
            value = config.get(account_name, entry)

        if not value:
            if entry in ['secret_access_key', 'access_key_id']:
                try:
                    value = get_account_info_from_aws(account, entry)
                except EC2AccountException:
                    msg = 'Unable to determine the %s value from '
                    msg += '~/.ec2utils.conf, ~/.aws/config, or '
                    msg += '~/.aws/credentials for account/profile %s.'
                    raise EC2AccountException(msg % (entry, account))
            else:
                msg = 'Unable to get %s value from account/region section %s'
                raise EC2AccountException(msg % (entry, account))

    return value


# -----------------------------------------------------------------------------
def get_regions(command_args, access_key, secret_key):
    """Return a list of connected regions if no regions are specified
       on the command line"""
    regions = None
    if command_args.regions:
        regions = command_args.regions.split(',')
    else:
        session = boto3.session.Session(
            aws_access_key_id=access_key,
            aws_secret_access_key=secret_key
        )
        sts_client = session.client('sts')

        try:
            # arn format is always "arn:{partition_name}"
            partition = sts_client.get_caller_identity()['Arn'].split(':')[1]
        except (KeyError, IndexError):
            partition = 'aws'  # default to public aws partition

        regions = session.get_available_regions(
            'ec2',
            partition_name=partition
        )

    return regions


# -----------------------------------------------------------------------------
def _basic_account_check(config, command_args):
    """Basic check for account presense."""
    account = command_args.accountName
    if not account:
        msg = 'Cannot determine ssh key, no account given and not specified '
        msg += 'on command line'
        raise EC2AccountException(msg)
    acctName = generate_config_account_name(account)
    if not config.has_section(acctName):
        msg = 'Could not find account %s in configuration file' % acctName
        raise EC2AccountException(msg)
    return 1


# ----------------------------------------------------------------------------
def _no_name_warning(image, log_callback):
    """Print a warning for images that have no name"""
    msg = 'WARNING: Found image with no name, ignoring for search results. '
    msg += 'Image ID: %s' % image['ImageId']
    log_callback.info(msg)


# ----------------------------------------------------------------------------
def validate_account_numbers(share_with):
    accounts = list(filter(None, share_with.split(',')))
    if accounts:
        return all(map(re.match, repeat(r'^\d{12}$'), accounts))
    return False


# ----------------------------------------------------------------------------
def get_logger(verbose):
    """
    Return new console logger at provided log level.
    """
    if verbose:
        log_level = logging.DEBUG
    else:
        log_level = logging.INFO

    logger = logging.getLogger('ec2imgutils')
    logger.setLevel(log_level)

    console_handler = logging.StreamHandler()
    console_handler.setLevel(log_level)
    console_handler.setFormatter(logging.Formatter('%(message)s'))

    logger.addHandler(console_handler)
    return logger


# ----------------------------------------------------------------------------
def get_version():
    version_file_name = 'VERSION'
    base_path = os.path.dirname(__file__)
    version = open(base_path + os.sep + version_file_name, 'r').read()
    return version.rstrip('\n')



# Copyright 2020 SUSE LLC
#
# This file is part of ec2imgutils
#
# ec2imgutils is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ec2imgutils is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ec2pub

import logging
import pprint

import ec2imgutils.ec2utils as utils
from ec2imgutils.ec2imgutils import EC2ImgUtils
from ec2imgutils.ec2imgutilsExceptions import EC2ListImgException


class EC2ListImage(EC2ImgUtils):
    """List owned images in an account."""

    # --------------------------------------------------------------------
    def __init__(
            self,
            access_key=None,
            image_id=None,
            image_name=None,
            image_name_fragment=None,
            image_name_match=None,
            indent=0,
            secret_key=None,
            log_level=logging.INFO,
            log_callback=None,
            verbose=0
    ):
        EC2ImgUtils.__init__(
            self,
            log_level=log_level,
            log_callback=log_callback
        )

        self.access_key = access_key
        self.indent = indent
        self.image_id = image_id
        self.image_name = image_name
        self.image_name_fragment = image_name_fragment
        self.image_name_match = image_name_match
        self.secret_key = secret_key
        self.verbose = verbose

    # ---------------------------------------------------------------------
    def list_images(self):
        """List images that meet the criteria"""
        owned_images = self._get_owned_images()
        if self.image_id:
            return utils.find_images_by_id(
                owned_images, self.image_id
            )
        elif self.image_name:
            return utils.find_images_by_name(
                owned_images, self.image_name, self.log
            )
        elif self.image_name_fragment:
            return utils.find_images_by_name_fragment(
                owned_images, self.image_name_fragment, self.log
            )
        elif self.image_name_match:
            try:
                return utils.find_images_by_name_regex_match(
                    owned_images, self.image_name_match, self.log
                )
            except Exception:
                msg = 'Unable to complie regular expression "%s"'
                msg = msg % self.image_name_match
                raise EC2ListImgException(msg)
        else:
            return owned_images

    # ---------------------------------------------------------------------
    def output_image_list(self):
        """Output the images that match in the account"""
        self._connect()
        images = self.list_images()

        for image in images:
            output = ' ' * self.indent
            if self.verbose == 0:
                self.log.info(output + image.get('Name'))
            elif self.verbose == 1:
                self.log.info(
                    output + image.get('Name') + '\t' + image.get('ImageId')
                )
            else:
                pp = pprint.PrettyPrinter(indent=(4 + self.indent))
                self.log.info(pp.pformat(image))
                if len(images) > 1 and image != images[-1]:
                    self.log.info('')

    # ---------------------------------------------------------------------
    def set_indent(self, indent):
        """Set the indent level for the output"""
        self.indent = indent



#!/usr/bin/python3
#
# Copyright (c) 2015 SUSE Linux GmbH.  All rights reserved.
#
# This file is part of ec2utils
#
# ec2utils is free software: you can redistribute it
# and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# ec2utils is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ec2utils. If not, see
# <http://www.gnu.org/licenses/>.
#

import logging
import os
import pytest

from ec2imgutils import ec2utils
from ec2imgutils.ec2imgutilsExceptions import (
    EC2AccountException,
    EC2ConfigFileParseException,
)

this_path = os.path.dirname(os.path.abspath(__file__))
data_path = this_path + os.sep + 'data'

logger = logging.getLogger('ec2imgutils')
logger.setLevel(logging.INFO)


class Turncoat:
    pass


def test_find_images_by_id_find_one():
    """Test find_images_by_id finds an image"""
    images = _get_test_images()
    found_images = ec2utils.find_images_by_id(images, 1)
    assert 1 == len(found_images)
    assert 1 == found_images[0]['ImageId']


def test_find_images_by_id_find_none():
    """Test find_images_by_id finds nothing and does not error"""
    images = _get_test_images()
    found_images = ec2utils.find_images_by_id(images, 5)
    assert 0 == len(found_images)


def test_find_images_by_name_find_some():
    """Test find_images_by_name finds images"""
    images = _get_test_images()
    image = {}
    image['Name'] = 'testimage-1'
    images.append(image)
    found_images = ec2utils.find_images_by_name(images, 'testimage-1', logger)
    assert 2 == len(found_images)
    assert 'testimage-1' == found_images[1]['Name']


def test_find_images_by_name_pending_image():
    """
    Test find_images_by_name does not raise if an image is pending.

    If Status is pending image will not have a name key.
    """
    images = _get_test_images()
    image = {}
    image['Status'] = 'pending'
    image['ImageId'] = 'testimage-3'
    images.append(image)
    ec2utils.find_images_by_name(images, 'testimage-1', logger)


def test_find_images_by_name_find_none():
    """Test find_images_by_name finds nothing and does not error"""
    images = _get_test_images()
    found_images = ec2utils.find_images_by_name(images, 'test', logger)
    assert 0 == len(found_images)


def test_find_images_by_name_fragment_find_some():
    """Test find_images_by_name_fragment finds images"""
    images = []
    image = {}
    image['Name'] = 'this-is-different'
    images.append(image)
    image = {}
    image['Name'] = 'find-this'
    images.append(image)
    image = {}
    image['Name'] = 'testimg'
    images.append(image)
    found_images = ec2utils.find_images_by_name_fragment(
        images,
        'this',
        logger
    )
    assert 2 == len(found_images)


def test_find_images_by_name_fragment_find_none():
    """Test find_images_by_name_fragment finds nothing and does not error"""
    images = _get_test_images()
    found_images = ec2utils.find_images_by_name_fragment(
        images,
        'foo',
        logger
    )
    assert 0 == len(found_images)


def test_find_images_by_name_regex_match_find_some():
    """Test find_images_by_name_regex_match finds images"""
    images = _get_test_images()
    found_images = ec2utils.find_images_by_name_regex_match(
        images,
        '^test',
        logger
    )
    assert 2 == len(found_images)


def test_find_images_by_name_regex_match_find_none():
    """Test find_images_by_name_regex_match finds  nothing and
       does not error"""
    images = _get_test_images()
    found_images = ec2utils.find_images_by_name_regex_match(
        images,
        '^-1',
        logger
    )
    assert 0 == len(found_images)


def test_find_images_by_name_regex_match_invalid_re():
    """Test find_images_by_name_regex_match throws with invalid expression"""
    images = _get_test_images()

    with pytest.raises(Exception):
        ec2utils.find_images_by_name_regex_match(
            images,
            't*{2}',
            logger
        )


def test_generate_config_account_name():
    """Test generate_config_account_name returns the expected name"""
    expected = 'account-foo'
    account_name = ec2utils.generate_config_account_name('foo')
    assert expected == account_name


def test_generate_config_region_name():
    """Test generate_config_region_name returns the expected name"""
    expected = 'region-bar'
    region_name = ec2utils.generate_config_region_name('bar')
    assert expected == region_name


def test_get_config_invalid():
    """Test get_config with an invalid configuration file"""
    config_file = data_path + os.sep + 'invalid.cfg'

    with pytest.raises(EC2ConfigFileParseException):
        ec2utils.get_config(config_file)


def test_get_from_config_from_account():
    """Test get_from_config returns expected data form an account"""
    config_file = data_path + os.sep + 'complete.cfg'
    config = ec2utils.get_config(config_file)
    expected = 'AAAAAAAAAAAAAA'
    access_key_id = ec2utils.get_from_config(
        'tester',
        config,
        None,
        'access_key_id',
        '--access-id')
    assert expected == access_key_id


def test_get_from_config_region_override():
    """Test get_from_config returns expected data if account setting
       is overridden ina a region"""
    config_file = data_path + os.sep + 'complete.cfg'
    config = ec2utils.get_config(config_file)
    expected = 'east-region'
    ssh_key_name = ec2utils.get_from_config(
        'tester',
        config,
        'us-east-1',
        'ssh_key_name',
        '--ssh-key-pair')
    assert expected == ssh_key_name


def test_get_from_config_invalid_account_name():
    """Test get_from_config throws if an invalid account name is given"""
    config_file = data_path + os.sep + 'complete.cfg'
    config = ec2utils.get_config(config_file)

    with pytest.raises(EC2AccountException):
        ec2utils.get_from_config(
            'foo',
            config,
            None,
            'ssh_key_name',
            '--ssh-key-pair')


def test_get_from_config_no_account_name():
    """Test get_from_config throws if no account name is given"""
    config_file = data_path + os.sep + 'complete.cfg'
    config = ec2utils.get_config(config_file)

    with pytest.raises(EC2AccountException):
        ec2utils.get_from_config(
            None,
            config,
            None,
            'ssh_key_name',
            '--ssh-key-pair')


def test_get_regions_from_cmd():
    """Test get_regions returns expected result for given command"""
    command_args = Turncoat()
    command_args.regions = 'milk,cookies,chocolate'
    expected = command_args.regions.split(',')
    regions = ec2utils.get_regions(command_args, '123', '456')
    assert expected == regions


# --------------------------------------------------------------------
# Helpers
def _get_test_images():
    images = []
    base_name = 'testimage-'
    for cnt in range(2):
        image = {}
        image['ImageId'] = cnt
        image['Name'] = base_name + '%d' % cnt
        images.append(image)

    return images



# Copyright (c) 2020 SUSE LLC
#
# This file is part of ec2imgutils.
#
# ec2imgutils is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ec2imgutils is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ec2imgutils.  If not, see <http://www.gnu.org/licenses/>.


class EC2AccountException(Exception):
    pass


class EC2ConfigFileParseException(Exception):
    pass


class EC2ConnectionException(Exception):
    pass


class EC2DeprecateImgException(Exception):
    pass


class EC2ListImgException(Exception):
    pass


class EC2PublishImgException(Exception):
    pass


class EC2RemoveImgException(Exception):
    pass


class EC2UploadImgException(Exception):
    pass



# Copyright 2018 SUSE LLC
#
# This file is part of ec2imgutils
#
# eec2imgutils is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ec2imgutils is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ec2uploadimg. If not, see <http://www.gnu.org/licenses/>.

import json
import logging
import os
import paramiko
import sys
import threading
import time


from ec2imgutils.ec2imgutils import EC2ImgUtils
from ec2imgutils.ec2imgutilsExceptions import EC2UploadImgException


class EC2ImageUploader(EC2ImgUtils):
    """Upload the given image to Amazon EC2"""

    def __init__(self,
                 access_key=None,
                 backing_store='ssd',
                 billing_codes=None,
                 bootkernel=None,
                 config=None,
                 ena_support=False,
                 image_arch='x86_64',
                 image_description='AWS EC2 AMI',
                 image_name='default',
                 image_virt_type='hvm',
                 inst_user_name=None,
                 launch_ami=None,
                 launch_inst_type='m1.small',
                 region=None,
                 root_volume_size=10,
                 running_id=None,
                 secret_key=None,
                 security_group_ids='',
                 session_token=None,
                 sriov_type=None,
                 ssh_key_pair_name=None,
                 ssh_key_private_key_file=None,
                 ssh_timeout=300,
                 use_grub2=False,
                 use_private_ip=False,
                 vpc_subnet_id='',
                 wait_count=1,
                 log_level=logging.INFO,
                 log_callback=None
                 ):
        EC2ImgUtils.__init__(
            self,
            log_level=log_level,
            log_callback=log_callback
        )

        self.access_key = access_key
        self.backing_store = backing_store
        self.billing_codes = billing_codes
        self.bootkernel = bootkernel
        self.ena_support = ena_support
        self.image_arch = image_arch
        self.image_description = image_description
        self.image_name = image_name
        self.image_virt_type = image_virt_type
        self.inst_user_name = inst_user_name
        self.launch_ami_id = launch_ami
        self.launch_ins_type = launch_inst_type
        self.region = region
        self.root_volume_size = int(root_volume_size)
        self.running_id = running_id
        self.secret_key = secret_key
        self.security_group_ids = security_group_ids
        self.session_token = session_token
        self.sriov_type = sriov_type
        self.ssh_key_pair_name = ssh_key_pair_name
        self.ssh_key_private_key_file = ssh_key_private_key_file
        self.ssh_timeout = ssh_timeout
        self.use_grub2 = use_grub2
        self.use_private_ip = use_private_ip
        self.vpc_subnet_id = vpc_subnet_id
        self.wait_count = wait_count

        self.created_volumes = []
        self.default_sleep = 10
        self.device_ids = ['f', 'g', 'h', 'i', 'j']
        self.instance_ids = []
        if self.running_id:
            self.instance_ids.append(self.running_id)
        self.next_device_id = 0
        self.operation_complete = False
        self.percent_transferred = 0
        self.ssh_client = None
        self.storage_volume_size = 2 * self.root_volume_size
        self.aborted = False

    def abort(self):
        """
        Set the abort flag to take appropriate action and stop image creation.
        """
        self.log.debug(
            "Aborted upload, please wait while AWS resources get cleaned"
            " up. This may take a few minutes!"
        )
        self.aborted = True

    # ---------------------------------------------------------------------
    def _attach_volume(self, volume, device=None):
        """Attach the given volume to the given instance"""
        if self.aborted:
            return
        if not device:
            device = self._get_next_disk_id()
        self._connect().attach_volume(
                VolumeId=volume['VolumeId'],
                InstanceId=self.helper_instance['InstanceId'],
                Device=device)
        self.log.debug('Wait for volume attachment')
        self.operation_complete = False
        self._show_progress()
        waiter = self._connect().get_waiter('volume_in_use')
        repeat_count = 1
        error_msg = 'Unable to attach volume'
        while repeat_count <= self.wait_count:
            try:
                wait_status = waiter.wait(
                    VolumeIds=[volume['VolumeId']],
                    Filters=[
                        {
                            'Name': 'attachment.status',
                            'Values': ['attached']
                        }
                    ]
                )
            except Exception:
                wait_status = 1
            if self.log_level == logging.DEBUG:
                self.progress_timer.cancel()
            repeat_count = self._check_wait_status(
                wait_status,
                error_msg,
                repeat_count
            )

    # ---------------------------------------------------------------------
    def _change_mount_point_permissions(self, target, permissions):
        """Change the permissions of the given target to the given value"""
        command = 'chmod %s %s' % (permissions, target)
        result = self._execute_ssh_command(command)

        return result

    # ---------------------------------------------------------------------
    def _check_image_exists(self):
        """Check if an image with the given name already exists"""
        my_images = self._get_owned_images()
        for image in my_images:
            if image['Name'] == self.image_name:
                msg = 'Image with name "%s" already exists' % self.image_name
                raise EC2UploadImgException(msg)

    # ---------------------------------------------------------------------
    def _check_security_groups_exist(self):
        """Check that the specified security groups exist"""
        try:
            self._connect().describe_security_groups(
                GroupIds=self.security_group_ids.split(',')
            )
        except Exception:
            error_msg = 'One or more of the specified security groups '
            error_msg += 'could not be found: %s' % self.security_group_ids
            raise EC2UploadImgException(error_msg)

    # ---------------------------------------------------------------------
    def _check_subnet_exists(self):
        """Verify that the subnet being used for the helper instance
           exists"""
        try:
            self._connect().describe_subnets(SubnetIds=[self.vpc_subnet_id])
        except Exception:
            error_msg = 'Specified subnet %s not found' % self.vpc_subnet_id
            raise EC2UploadImgException(error_msg)

    # ---------------------------------------------------------------------
    def _check_virt_type_consistent(self):
        """When using root swap the virtualization type of the helper
           image and the target image must be the same"""
        if self.launch_ami_id:
            image = self._connect().describe_images(
                ImageIds=[self.launch_ami_id]
            )['Images'][0]
        elif self.running_id:
            helper_instance = self._get_helper_instance()
            image = self._connect().describe_images(
                ImageIds=[helper_instance['ImageId']]
            )['Images'][0]
        else:
            error_msg = 'Could not determine helper image virtualization '
            error_msg += 'type necessary for root swap method'
            raise EC2UploadImgException(error_msg)

        if not self.image_virt_type == image['VirtualizationType']:
            error_msg = 'Virtualization type of the helper image and the '
            error_msg += 'target image must be the same when using '
            error_msg += 'root-swap method for image creation.'
            raise EC2UploadImgException(error_msg)

    # ---------------------------------------------------------------------
    def _check_wait_status(
            self,
            wait_status,
            error_msg,
            repeat_count,
            skip_cleanup=False
    ):
        """Check the wait status form the waiter and take appropriate action"""
        if wait_status:
            if self.log_level == logging.DEBUG:
                print()
            if repeat_count == self.wait_count:
                self.operation_complete = True
                if self.log_level == logging.DEBUG:
                    self.progress_timer.cancel()
                time.sleep(self.default_sleep)  # Wait for the thread
                if not skip_cleanup:
                    self._clean_up()
                raise EC2UploadImgException(error_msg)
            repeat_count += 1
            self.log.info('Entering wait loop number %d of %d' % (
                repeat_count,
                self.wait_count
            ))
            self.operation_complete = False
            self._show_progress()
        else:
            repeat_count = self.wait_count + 1
            self.operation_complete = True
            if self.log_level == logging.DEBUG:
                self.progress_timer.cancel()
            time.sleep(self.default_sleep)  # Wait for the thread
            if self.log_level == logging.DEBUG:
                print()

        return repeat_count

    # ---------------------------------------------------------------------
    def _clean_up(self):
        """Clean up the given resources"""
        if self.ssh_client:
            self.ssh_client.close()
        if self.instance_ids:
            self._connect().terminate_instances(InstanceIds=self.instance_ids)
            waiter = self._connect().get_waiter('instance_terminated')
            repeat_count = 1
            error_msg = 'Instance did not stop within allotted time'
            while repeat_count <= self.wait_count:
                try:
                    wait_status = waiter.wait(
                        InstanceIds=[self.helper_instance['InstanceId']],
                        Filters=[
                            {
                                'Name': 'instance-state-name',
                                'Values': ['terminated']
                            }
                        ]
                    )
                except Exception:
                    wait_status = 1
                if self.log_level == logging.DEBUG:
                    self.progress_timer.cancel()
                repeat_count = self._check_wait_status(
                    wait_status,
                    error_msg,
                    repeat_count
                )
        if self.created_volumes:
            for volume in self.created_volumes:
                self._detach_volume(volume)
                self._remove_volume(volume)

        self.created_volumes = []
        self.instance_ids = []

    # ---------------------------------------------------------------------
    def _create_block_device_map(self, snapshot):
        """Create a block device map with the given snapshot"""
        if self.aborted:
            return
        # We assume the root image has 1 partition (either case)
        root_device_name = self._determine_root_device()

        if self.backing_store == 'mag':
            backing_store = 'standard'
        elif self.backing_store == 'ssd':
            backing_store = 'gp2'
        else:
            backing_store = self.backing_store

        block_device_map = {
            'DeviceName': root_device_name,
            'Ebs': {
                'SnapshotId': snapshot['SnapshotId'],
                'VolumeSize': self.root_volume_size,
                'DeleteOnTermination': True,
                'VolumeType': backing_store
            }
        }

        return [block_device_map]

    # ---------------------------------------------------------------------
    def _create_image_root_volume(self, source):
        if self.aborted:
            return
        """Create a root volume from the image"""
        self._check_image_exists()
        if self.vpc_subnet_id:
            self._check_subnet_exists()
        if self.security_group_ids:
            self._check_security_groups_exist()
        if self.running_id:
            # When using an already running instance, simply look up this one
            helper_instance = self._get_helper_instance()
        else:
            helper_instance = self._launch_helper_instance()
        self.helper_instance = helper_instance
        store_volume = self._create_storge_volume()
        self._attach_volume(store_volume)
        self._establish_ssh_connection()
        store_device_id = self._find_device_name(self.storage_volume_size)
        target_root_volume = self._create_target_root_volume()
        self._attach_volume(target_root_volume)
        target_root_device_id = self._find_device_name(self.root_volume_size)
        self._format_storage_volume(store_device_id)
        self._create_storage_filesystem(store_device_id)
        mount_point = self._mount_storage_volume(store_device_id)
        self._change_mount_point_permissions(mount_point, '777')
        image_filename = self._upload_image(mount_point, source)
        raw_image_filename = self._unpack_image(mount_point, image_filename)
        self._dump_root_fs(
            mount_point,
            raw_image_filename,
            target_root_device_id
        )
        self._execute_ssh_command('umount %s' % mount_point)
        self._end_ssh_session()
        self._detach_volume(target_root_volume)
        self._detach_volume(store_volume)

        return target_root_volume

    # ---------------------------------------------------------------------
    def _create_snapshot(self, volume):
        """Create a snapshot from a volume"""
        if self.aborted:
            return
        snapshot = self._connect().create_snapshot(
            VolumeId=volume['VolumeId'],
            Description=self.image_description
        )
        self.log.debug(
            'Waiting for snapshot creation: {}'.format(snapshot['SnapshotId'])
        )
        self.operation_complete = False
        self._show_progress()
        waiter = self._connect().get_waiter('snapshot_completed')
        repeat_count = 1
        error_msg = 'Unable to create snapshot'
        while repeat_count <= self.wait_count:
            try:
                wait_status = waiter.wait(
                    SnapshotIds=[snapshot['SnapshotId']],
                    Filters=[
                        {
                            'Name': 'status',
                            'Values': ['completed']
                        }
                    ]
                )
            except Exception:
                wait_status = 1
            if self.log_level == logging.DEBUG:
                self.progress_timer.cancel()
            repeat_count = self._check_wait_status(
                wait_status,
                error_msg,
                repeat_count
            )

        return snapshot

    # ---------------------------------------------------------------------
    def _create_storage_filesystem(self, device_id):
        """Create an ext3 filesystem on the storage volume"""
        self.log.debug('Creating ext3 filesystem on storage volume')
        filesystem_partition = '%s1' % device_id
        command = 'mkfs -t ext3 %s' % filesystem_partition
        result = self._execute_ssh_command(command)

        return result

    # ---------------------------------------------------------------------
    def _create_storge_volume(self):
        """Create the volume that will be used to store the image before
           dumping it to the new root volume"""
        return self._create_volume('%s' % self.storage_volume_size)

    # ---------------------------------------------------------------------
    def _create_target_root_volume(self):
        """Create the volume that will be used as the root volume for
           the image we are creating."""
        return self._create_volume('%s' % self.root_volume_size)

    # ---------------------------------------------------------------------
    def _create_volume(self, size):
        """Create a volume"""
        if self.aborted:
            return
        volume = self._connect().create_volume(
            Size=int(size),
            AvailabilityZone=self.zone,
            VolumeType='gp2'
        )
        self.log.debug(
            'Waiting for volume creation: {}'.format(volume['VolumeId'])
        )
        self.operation_complete = False
        self._show_progress()
        waiter = self._connect().get_waiter('volume_available')
        repeat_count = 1
        error_msg = 'Time out for Volume creation reached, '
        error_msg += 'terminating instance and deleting volume'
        while repeat_count <= self.wait_count:
            try:
                wait_status = waiter.wait(
                    VolumeIds=[volume['VolumeId']],
                    Filters=[
                        {
                            'Name': 'status',
                            'Values': ['available']
                        }
                    ]
                )
            except Exception:
                wait_status = 1
            if self.log_level == logging.DEBUG:
                self.progress_timer.cancel()
            repeat_count = self._check_wait_status(
                wait_status,
                error_msg,
                repeat_count
            )

        self.created_volumes.append(volume)

        return volume

    # ---------------------------------------------------------------------
    def _detach_volume(self, volume):
        """Detach the given volume"""
        if not volume:
            return
        volume = self._connect().describe_volumes(
            VolumeIds=[volume['VolumeId']])['Volumes'][0]
        if volume['State'] == 'available':
            # Not attached, nothing to do
            return 1

        self._connect().detach_volume(VolumeId=volume['VolumeId'])
        self.log.debug('Wait for volume to detach')
        self.operation_complete = False
        self._show_progress()
        waiter = self._connect().get_waiter('volume_available')
        repeat_count = 1
        error_msg = 'Unable to detach volume'
        while repeat_count <= self.wait_count:
            try:
                wait_status = waiter.wait(
                    VolumeIds=[volume['VolumeId']],
                    Filters=[
                        {
                            'Name': 'status',
                            'Values': ['available']
                        }
                    ]
                )
            except Exception:
                wait_status = 1
            if self.log_level == logging.DEBUG:
                self.progress_timer.cancel()
            repeat_count = self._check_wait_status(
                wait_status,
                error_msg,
                repeat_count
            )

        return 1

    # ---------------------------------------------------------------------
    def _determine_root_device(self):
        """Figure out what the root device should be"""
        root_device_name = '/dev/sda'
        if self.image_virt_type == 'hvm':
            root_device_name = '/dev/sda1'

        return root_device_name

    # ---------------------------------------------------------------------
    def _device_exists(self, device_id):
        """Verify that the device node can be found on the remote system"""
        command = 'ls %s' % device_id
        result = self._execute_ssh_command(command)

        return device_id == result

    # ---------------------------------------------------------------------
    def _dump_root_fs(self, image_dir, raw_image_name, target_root_device):
        """Dump the raw image to the target device"""
        if self.aborted:
            return
        self.log.debug('Dumping raw image to new target root volume')
        command = 'dd if=%s/%s of=%s bs=32k' % (image_dir,
                                                raw_image_name,
                                                target_root_device)
        result = self._execute_ssh_command(command)

        return result

    # ---------------------------------------------------------------------
    def _end_ssh_session(self):
        """End the SSH session"""
        if not self.ssh_client:
            return

        self.ssh_client.close()
        del self.ssh_client
        self.ssh_client = None

        return 1

    # ---------------------------------------------------------------------
    def _establish_ssh_connection(self):
        """Connect to the running instance with ssh"""
        if self.aborted:
            return
        self.log.debug('Waiting to obtain instance IP address')
        instance_ip = self.helper_instance.get('PublicIpAddress')
        if self.use_private_ip:
            instance_ip = self.helper_instance.get('PrivateIpAddress')
        timeout_counter = 1
        while not instance_ip:
            instance = self._connect().describe_instances(
                InstanceIds=[self.helper_instance['InstanceId']]
            )['Reservations'][0]['Instances'][0]
            instance_ip = instance.get('PublicIpAddress')
            if self.use_private_ip:
                instance_ip = instance.get('PrivateIpAddress')
            if self.log_level == logging.DEBUG:
                print('. ', end=' ')
                sys.stdout.flush()
            if timeout_counter * self.default_sleep >= self.ssh_timeout:
                msg = 'Unable to obtain the instance IP address'
                raise EC2UploadImgException(msg)
            timeout_counter += 1
        if self.log_level == logging.DEBUG:
            print()
        client = paramiko.client.SSHClient()
        client.set_missing_host_key_policy(paramiko.WarningPolicy())
        self.log.debug('Attempt ssh connection to %s' % instance_ip)
        ssh_connection = None
        timeout_counter = 1
        while not ssh_connection:
            try:
                ssh_connection = client.connect(
                    key_filename=self.ssh_key_private_key_file,
                    username=self.inst_user_name,
                    hostname=instance_ip
                )
            except (Exception, ConnectionResetError):
                if self.log_level == logging.DEBUG:
                    print('. ', end=' ')
                    sys.stdout.flush()
                time.sleep(self.default_sleep)
                if (
                        timeout_counter * self.default_sleep >=
                        self.ssh_timeout):
                    self._clean_up()
                    msg = 'Time out for ssh connection reached, '
                    msg += 'could not connect'
                    raise EC2UploadImgException(msg)
                timeout_counter += 1
            else:
                ssh_connection = True
                print()

        self.ssh_client = client

    # ---------------------------------------------------------------------
    def _execute_ssh_command(self, command):
        """Execute a command on the remote machine, on error raise an exception
           return the result of stdout"""
        if self.aborted:
            return
        if self.inst_user_name != 'root':
            command = 'sudo %s' % command

        if not self.ssh_client:
            msg = 'No ssh connection established, cannot execute command'
            raise EC2UploadImgException(msg)
        stdin, stdout, stderr = self.ssh_client.exec_command(command,
                                                             get_pty=True)
        cmd_error = stderr.read()
        if cmd_error:
            self._clean_up()
            msg = 'Execution of "%s" failed with the following error' % command
            msg += '\n%s' % cmd_error
            raise EC2UploadImgException(msg)

        return stdout.read().strip().decode('utf-8')

    # ---------------------------------------------------------------------
    def _find_device_name(self, device_size):
        """Match an attached volume by size"""
        try:
            lsblk_out = json.loads(self._execute_ssh_command('lsblk -a -J'))
        except Exception:
            self.log.error(
                '"lsblk -a -J" command failed on helper instance. Ensure the '
                'helper instance has lsblk >= 2.27 which has the json option.'
            )
            lsblk_out = {'blockdevices': tuple()}

        for device in lsblk_out['blockdevices']:
            if device.get('children'):
                continue
            this_device_size = device['size']
            unit = this_device_size[-1]
            try:
                size = int(this_device_size[:-1])
            except ValueError:
                self.log.info('Skipping non integer sized disk')
                continue

            size_multiplier = 1
            if unit == 'T':
                size_multiplier = 1024
            disk_size = size * size_multiplier
            if disk_size == device_size:
                # We do not need to worry about finding the same device twice
                # Only 2 disks get attached and they are known to have
                # different sizes
                return '/dev/%s' % device['name']

        self._clean_up()
        msg = 'Could not find attached disk device with size %sG' % device_size
        raise EC2UploadImgException(msg)

    # ---------------------------------------------------------------------
    def _format_storage_volume(self, device_id):
        """Format the storage volume"""
        if self.aborted:
            return
        self.log.debug('Formating storage volume')
        parted = self._get_command_from_instance('parted')
        sfdisk = None
        if not parted:
            sfdisk = self._get_command_from_instance('sfdisk')

        if not parted and not sfdisk:
            self._clean_up()
            msg = 'Neither parted nor sfdisk found on target image. '
            msg += 'Need to partition storage device but cannot, exiting.'
            raise EC2UploadImgException(msg)

        if parted:
            command = '%s -s %s mklabel gpt' % (parted, device_id)
            result = self._execute_ssh_command(command)
            blockdev = self._get_command_from_instance('blockdev')
            command = '%s --getsize %s' % (blockdev, device_id)
            size = self._execute_ssh_command(command)
            command = ('%s -s %s unit s mkpart primary 2048 %d' %
                       (parted, device_id, int(size)-100))
            result = self._execute_ssh_command(command)
        else:
            command = 'echo ",,L" > /tmp/partition.txt'
            result = self._execute_ssh_command(command)
            command = '%s %s < /tmp/partition.txt' % (sfdisk, device_id)
            result = self._execute_ssh_command(command)

        return result

    # ---------------------------------------------------------------------
    def _get_command_from_instance(self, command):
        """Get the location of the given command from the instance"""
        if self.aborted:
            return
        loc_cmd = 'which %s' % command
        location = self._execute_ssh_command(loc_cmd)

        if location.find('which: no') != -1:
            location = ''

        return location

    # ---------------------------------------------------------------------
    def _get_helper_instance(self):
        """Returns handle to running instance"""
        helper_instance = self._connect().describe_instances(
            InstanceIds=[self.running_id])['Reservations'][0]['Instances'][0]
        if helper_instance['State']['Name'] != 'running':
            msg = 'Helper instance %s is not running' % self.running_id
            raise EC2UploadImgException(msg)

        self.vpc_subnet_id = helper_instance['SubnetId']
        self._set_zone_to_use()

        return helper_instance

    # ---------------------------------------------------------------------
    def _get_next_disk_id(self):
        """Return the next device name for a storage volume"""
        device = '/dev/sd' + self.device_ids[self.next_device_id]
        self.next_device_id += 1

        return device

    # ---------------------------------------------------------------------
    def _launch_helper_instance(self):
        """Launch the helper instance that is used to create the new image"""
        if self.aborted:
            return
        self._set_zone_to_use()
        if self.security_group_ids:
            instance = self._connect().run_instances(
                ImageId=self.launch_ami_id,
                MinCount=1,
                MaxCount=1,
                KeyName=self.ssh_key_pair_name,
                InstanceType=self.launch_ins_type,
                Placement={'AvailabilityZone': self.zone},
                NetworkInterfaces=[
                    {
                        'DeviceIndex': 0,
                        'AssociatePublicIpAddress': not self.use_private_ip,
                        'SubnetId': self.vpc_subnet_id,
                        'Groups': self.security_group_ids.split(',')
                    }
                ],
                TagSpecifications=[
                    {
                        'ResourceType': 'instance',
                        'Tags': [
                            {
                                'Key': 'Name',
                                'Value': 'ec2uploadimg-helper-instance'
                            }
                        ]
                    }
                ]
            )['Instances'][0]
        else:
            instance = self._connect().run_instances(
                ImageId=self.launch_ami_id,
                MinCount=1,
                MaxCount=1,
                KeyName=self.ssh_key_pair_name,
                InstanceType=self.launch_ins_type,
                Placement={'AvailabilityZone': self.zone},
                NetworkInterfaces=[
                    {
                        'DeviceIndex': 0,
                        'AssociatePublicIpAddress': not self.use_private_ip,
                        'SubnetId': self.vpc_subnet_id
                    }
                ],
                TagSpecifications=[
                    {
                        'ResourceType': 'instance',
                        'Tags': [
                            {
                                'Key': 'Name',
                                'Value': 'ec2uploadimg-helper-instance'
                            }
                        ]
                    }
                ]
            )['Instances'][0]

        self.instance_ids.append(instance['InstanceId'])

        self.log.debug(
            'Waiting for instance: {}'.format(instance['InstanceId'])
        )
        self.operation_complete = False
        self._show_progress()
        waiter = self._connect().get_waiter('instance_running')
        repeat_count = 1
        error_msg = 'Time out for instance creation reached, '
        error_msg += 'terminating instance'
        while repeat_count <= self.wait_count:
            try:
                wait_status = waiter.wait(
                    InstanceIds=[instance['InstanceId']],
                    Filters=[
                        {
                            'Name': 'instance-state-name',
                            'Values': ['running']
                        }
                    ]
                )
            except Exception:
                wait_status = 1
            if self.log_level == logging.DEBUG:
                self.progress_timer.cancel()
            repeat_count = self._check_wait_status(
                wait_status,
                error_msg,
                repeat_count
            )

        return instance

    # ---------------------------------------------------------------------
    def _mount_storage_volume(self, device_id):
        """Mount the storage volume"""
        mount_point = '/mnt'
        mount_device = '%s1' % device_id
        command = 'mount %s %s' % (mount_device, mount_point)
        self._execute_ssh_command(command)

        return mount_point

    # ---------------------------------------------------------------------
    def _register_image(self, snapshot):
        """Register an image from the given snapshot"""
        if self.aborted:
            return
        block_device_map = self._create_block_device_map(snapshot)
        self.log.debug('Registering image')

        root_device_name = self._determine_root_device()
        register_args = {
            'Architecture': self.image_arch,
            'BlockDeviceMappings': block_device_map,
            'Description': self.image_description.strip(),
            'EnaSupport': self.ena_support,
            'Name': self.image_name,
            'RootDeviceName': root_device_name,
            'VirtualizationType': self.image_virt_type
        }
        if self.billing_codes:
            register_args['BillingProducts'] = self.billing_codes.split(',')
        if self.image_virt_type == 'paravirtual':
            register_args['KernelId'] = self.bootkernel
        if self.sriov_type:
            register_args['SriovNetSupport'] = self.sriov_type

        ami = self._connect().register_image(**register_args)

        return ami['ImageId']

    # ---------------------------------------------------------------------
    def _remove_volume(self, volume):
        """Delete the given volume from EC2"""
        if not volume:
            return
        self._connect().delete_volume(VolumeId=volume['VolumeId'])

        return 1

    # ---------------------------------------------------------------------
    def _set_zone_to_use(self):
        """Set the availability zone to use for all operations"""
        if self.vpc_subnet_id:
            # If a subnet is given we need to launch the helper instance
            # in the AZ wher ethe subnet is defined
            subnet = self._connect().describe_subnets(
                SubnetIds=[self.vpc_subnet_id]
            )['Subnets'][0]
            self.zone = subnet['AvailabilityZone']
            return
        zones = self._connect().describe_availability_zones()[
            'AvailabilityZones']
        availability_zones = []
        for zone in zones:
            availability_zones.append(zone['ZoneName'])
        self.zone = availability_zones[-1]

    # ---------------------------------------------------------------------
    def _show_progress(self, timeout_counter=1):
        """Progress indicator"""

        # Give EC2 some time to update it's state
        # Whenever _show_progress is called we get a waiter from EC2
        # The wait operation may fail if the EC2 state is not yet updated.
        # Taking a nap on the client side avoids the problem
        time.sleep(self.default_sleep)

        if self.log_level == logging.DEBUG:
            print('. ', end=' ')
            sys.stdout.flush()
            timeout_counter += 1
            if not self.operation_complete:
                self.progress_timer = threading.Timer(
                    self.default_sleep,
                    self._show_progress,
                    args=[timeout_counter]
                )
                self.progress_timer.start()

    # ---------------------------------------------------------------------
    def _upload_image(self, target_dir, source):
        """Upload the source file to the instance"""
        if self.aborted:
            return
        filename = source.split(os.sep)[-1]
        sftp = self.ssh_client.open_sftp()
        try:
            self.log.debug('Uploading image file: {}'.format(source))
            sftp.put(source,
                     '%s/%s' % (target_dir, filename),
                     self._upload_progress)
            if self.log_level == logging.DEBUG:
                print()
        except Exception as e:
            self._clean_up()
            raise e

        return filename

    # ---------------------------------------------------------------------
    def _upload_progress(self, transferred_bytes, total_bytes):
        """In verbose mode give an upload progress indicator"""
        if self.log_level == logging.DEBUG:
            percent_complete = (float(transferred_bytes) / total_bytes) * 100
            if percent_complete - self.percent_transferred >= 10:
                print('.', end=' ')
                sys.stdout.flush()
                self.percent_transferred = percent_complete

        return 1

    # ---------------------------------------------------------------------
    def _unpack_image(self, image_dir, image_filename):
        """Unpack the uploaded image file"""
        if self.aborted:
            return
        raw_image_file = None
        files = ''
        if (
                image_filename.find('.tar') != -1 or
                image_filename.find('.tbz') != -1 or
                image_filename.find('.tgz') != -1):
            command = 'tar -C %s -xvf %s/%s' % (image_dir,
                                                image_dir,
                                                image_filename)
            files = self._execute_ssh_command(command).split('\r\n')
        elif image_filename[-2:] == 'xz':
            files = [image_filename]
        elif image_filename[-3:] == 'raw':
            raw_image_file = image_filename

        if files:
            # Find the disk image
            for fl in files:
                if fl.strip()[-2:] == 'xz':
                    self.log.debug('Inflating image: {}'.format(fl))
                    command = 'xz -d %s/%s' % (image_dir, fl)
                    self._execute_ssh_command(command)
                    raw_image_file = fl.strip()[:-3]
                    break
                if fl.strip()[-4:] == '.raw':
                    raw_image_file = fl.strip()
                    break
        if not raw_image_file:
            self._clean_up()
            msg = 'Unable to find raw image file with .raw extension'
            raise EC2UploadImgException(msg)

        return raw_image_file

    # ---------------------------------------------------------------------
    def create_image(self, source):
        """Create an AMI (Amazon Machine Image) from the given source"""
        snapshot = self.create_snapshot(source)

        ami = self._register_image(snapshot)

        return ami

    # ---------------------------------------------------------------------
    def create_image_from_snapshot(self, source):
        """Create an AMI (Amazon Machine Image) from the given snapshot"""

        try:
            response = self._connect().describe_snapshots(SnapshotIds=[source])

        except Exception:
            self.log.error('Unable to retrieve details for snapshot %s',
                           source)
            sys.exit(1)

        snapshot = response['Snapshots'][0]
        ami = self._register_image(snapshot)

        return ami

    # ---------------------------------------------------------------------

    def create_image_use_root_swap(self, source):
        """Creae an AMI (Amazon Machine Image) from the given source using
           the root swap method"""

        self._check_virt_type_consistent()
        target_root_volume = self._create_image_root_volume(source)
        self._connect().stop_instances(InstanceIds=self.instance_ids)
        self.log.debug('Waiting for helper instance to stop')
        self.operation_complete = False
        self._show_progress()
        waiter = self._connect().get_waiter('instance_stopped')
        repeat_count = 1
        error_msg = 'Instance did not stop within allotted time'
        while repeat_count <= self.wait_count:
            try:
                wait_status = waiter.wait(
                    InstanceIds=[self.helper_instance['InstanceId']],
                    Filters=[
                        {
                            'Name': 'instance-state-name',
                            'Values': ['stopped']
                        }
                    ]
                )
            except Exception:
                wait_status = 1
            if self.log_level == logging.DEBUG:
                self.progress_timer.cancel()
            repeat_count = self._check_wait_status(
                wait_status,
                error_msg,
                repeat_count
            )

        # Find the current root volume
        my_volumes = self._connect().describe_volumes()['Volumes']
        current_root_volume = None
        device_id = None
        for volume in my_volumes:
            attached = volume['Attachments']
            if not attached:
                continue
            attached_to_instance = attached[0].get('InstanceId')
            if attached_to_instance == self.helper_instance['InstanceId']:
                current_root_volume = volume
                device_id = attached[0].get('Device')
                break

        self._detach_volume(current_root_volume)
        self._attach_volume(target_root_volume, device_id)
        self.log.debug('Creating new image')
        instance_info = self._connect().describe_instances(
            InstanceIds=[self.helper_instance['InstanceId']]
        )
        helper_instance = instance_info['Reservations'][0]['Instances']
        helper_dev_map = helper_instance[0]['BlockDeviceMappings'][0]
        if self.backing_store == 'mag':
            backing_store = 'standard'
        else:
            backing_store = 'gp2'
        block_device_map = {
            'DeviceName': helper_dev_map['DeviceName'],
            'Ebs': {
                'DeleteOnTermination':
                helper_dev_map['Ebs']['DeleteOnTermination'],
                'VolumeSize': self.root_volume_size,
                'VolumeType': backing_store
            }
        }
        ami = self._connect().create_image(
            BlockDeviceMappings=[block_device_map],
            InstanceId=self.helper_instance['InstanceId'],
            Name=self.image_name,
            Description=self.image_description,
            NoReboot=True
        )

        self.log.debug('Waiting for new image creation')
        self.operation_complete = False
        self._show_progress()
        waiter = self._connect().get_waiter('image_available')
        repeat_count = 1
        error_msg = 'Image creation did not complete within '
        error_msg += 'allotted time skipping clean up'
        while repeat_count <= self.wait_count:
            try:
                wait_status = waiter.wait(
                    ImageIds=[ami['ImageId']],
                    Filters=[
                        {
                            'Name': 'state',
                            'Values': ['available']
                        }
                    ]
                )
            except Exception:
                wait_status = 1
            if self.log_level == logging.DEBUG:
                self.progress_timer.cancel()
            repeat_count = self._check_wait_status(
                wait_status,
                error_msg,
                repeat_count,
                True
            )

        self._clean_up()
        return ami['ImageId']

    # ---------------------------------------------------------------------
    def create_snapshot(self, source):
        """Create a snapshot from the given source"""
        if self.log_level == logging.DEBUG:
            print()
        root_volume = self._create_image_root_volume(source)
        snapshot = self._create_snapshot(root_volume)
        self._clean_up()
        return snapshot
